/* This file was automatically generated.  Do not edit! */
#undef INTERFACE
void pclose2(int fdIn,FILE *pOut,int childPid);
NORETURN void fossil_panic(const char *zFormat,...);
void *fossil_utf8_to_unicode(const char *zUtf8);
int popen2(const char *zCmd,int *pfdIn,FILE **ppOut,int *pChildPid,int bDirect);
NORETURN void fossil_fatal(const char *zFormat,...);
